import time
from .formatter import Formatter, space
from .game import Game


class Session:
    """API for a session of gameplay of IQ Tester"""

    def __init__(self):
        self.f = Formatter(78)
        self.total_score = 0
        self.board_size = 5
        self.game = None
        self.played = 0
        self.keep_playing = True

    def average(self):
        if self.played == 0:
            return round(0, 1)
        return round(self.total_score / self.played, 1)

    @space
    def header(self):
        self.f.center('', fill='*', s=["BOLD", "BLUE"])
        self.f.center(' WELCOME TO IQ TESTER ', fill='*', s=["BOLD"])
        self.f.center('', fill='*', s=["BOLD", "BLUE"])

    @space
    def instructions(self):
        self.f.center("Start with any one hole empty.")
        self.f.center("As you jump the pegs remove them from the board.")
        self.f.center("Try to leave only one peg. See how you rate!.")

    @space
    def menu_options(self):
        """Display the main menu including statistics and gameplay options"""
        # menu width
        w = 40

        # menu header
        self.f.center('', fill='-', in_w=w - 2)
        self.f.center("", in_w=w, in_b='|')
        self.f.center("HOME MENU", ['BOLD'], in_w=w, in_b='|')
        self.f.center("", in_w=w, in_b='|')

        # game statistics
        msg = f"GAMES PLAYED: {self.played}"
        self.f.center(msg, ['BOLD', 'GREEN'], in_w=w, in_b='|')
        msg = f"YOUR TOTAL SCORE: {self.total_score}"
        self.f.center(msg, ['BOLD', 'GREEN'], in_w=w, in_b='|')
        msg = f"AVERAGE SCORE: {self.average()}"
        self.f.center(msg, ['BOLD', 'GREEN'], in_w=w, in_b='|')
        self.f.center("", in_w=w, in_b='|')

        # menu options
        self.f.center("New Game [ENTER]", ['BOLD', 'RED'], in_w=w, in_b='|')
        self.f.center("Edit Settings [s]", ['BOLD', 'RED'], in_w=w, in_b='|')
        self.f.center("Quit [q]", ['BOLD', 'RED'], in_w=w, in_b='|')

        # bottom border of menu
        self.f.center("", in_w=w, in_b='|')
        self.f.center('', fill='-', in_w=w - 2)

    def settings_menu(self):
        """Allow user to change certain gameplay settings"""
        # menu width
        w = 40

        # menu header
        self.f.center('', fill='-', in_w=w - 2)
        self.f.center("", in_w=w, in_b='|')
        self.f.center("SETTINGS MENU", ['BOLD'], in_w=w, in_b='|')
        self.f.center("", in_w=w, in_b='|')

        # menu options
        n = self.board_size
        self.f.center(f"Update Board Size [u]: {n}", in_w=w, in_b='|')
        self.f.center("Return to Main Menu [m]", in_w=w, in_b='|')

        # bottom border of menu
        self.f.center("", in_w=w, in_b='|')
        self.f.center('', fill='-', in_w=w - 2)

    @space
    def update_board_size(self):
        low = 4  # minimum board size (3 or less doesn't work)
        high = 6  # maximum board size (7 or more uses > 26 pegs)
        while True:
            n = self.f.prompt(f"Enter desired board size ({low} to {high}):")
            try:
                n = int(n)
                if low <= n <= high:
                    break
            except NameError:
                self.f.center("Board size must be an integer. Try again.")

        print()
        self.f.center(f"Updating board size to {n}...")
        self.board_size = n

    @space
    def footer(self):
        self.f.center("For even more fun compete with someone. Lots of luck!")
        self.f.center("Copyright (C) 1975 Venture MFG. Co., INC. U.S.A.")
        self.f.center("Python package `iqtester` by Andrew Tracey, 2022.")
        self.f.center("Follow me: https://www.github.com/andrewt110216")

    @space
    def main_selection(self):
        """Let user select a menu option"""
        play = self.f.prompt("PRESS ENTER FOR NEW GAME")
        return play

    @space
    def setting_selection(self):
        """Let user select an option from the settings menu"""
        setting = self.f.prompt("Make a selection from the above options")
        return setting

    def quit(self):
        """Handle user selection to quit playing"""
        self.f.center("Thanks for playing!", s=['BOLD'])
        self.footer()
        self.keep_playing = False

    def start(self):
        """Drive gameplay"""
        self.header()
        self.instructions()
        while self.keep_playing:
            self.menu_options()
            choice = self.main_selection().lower()
            if choice == "":
                self.game = Game(self.f, self.board_size)
                game_score = self.game.play()
                self.total_score += game_score
                self.played += 1
            elif choice == "s":
                self.settings_menu()
                setting_choice = self.setting_selection().lower()
                if setting_choice == "u":
                    self.update_board_size()
                    time.sleep(0.8)
                self.f.center("Returning to Main Menu...")
                time.sleep(1)
            else:
                self.quit()



from functools import wraps


def space(f):
    """Decorate class methods with a new line printed before execution"""

    @wraps(f)
    def wrapped(self, *args, **kwargs):
        print()
        res = f(self, *args, **kwargs)
        print()
        return res

    return wrapped


class Formatter:
    """An API to control formatting of the command line interface"""

    def __init__(self, width):
        self.w = width

    # style formatting options {'name': ('string code', length of string code)}
    formats = {
        "BLUE": ("\033[94m", 5),
        "GREEN": ("\033[92m", 5),
        "RED": ("\033[91m", 5),
        "BOLD": ("\033[1m", 4),
        "END": ("\033[0m", 4),
    }

    def center(self, msg, s=[], fill=" ", in_w=0, in_b="", fc=0, end="\n"):
        """Style text and center across width with optional inner container"""

        # n represents actual printed length of msg, less formatting characters
        n = len(msg) - fc

        # apply styles to message and inner container border
        fmt_start = ""
        for style in s:
            if style not in self.formats:
                raise NotImplementedError(f'Style "{style}" not implemented.')
            fmt_start += self.formats[style][0]
        msg = fmt_start + msg + self.formats["END"][0]

        # margin refers to combined space on each side of inner container
        margin = self.w - in_w if in_w else 0
        lmargin = margin // 2

        # padding refers to combined space on each side of msg inside container
        padding = self.w - margin - n - 2 * len(in_b)
        lpad = padding // 2
        # right padding only needed if fill is not ' ' or for inner container
        rpad = 0
        if in_w or fill != " ":
            rpad = lpad if padding % 2 == 0 else lpad + 1

        # print left margin, border, left padding, msg, right padding, border
        disp = " " * lmargin + in_b + fill * lpad + msg + fill * rpad + in_b
        print(disp, end=end)

    def apply(self, text, options=["BOLD"]):
        """Return string with formatting applied and number of format chars"""
        fmt_start = ""
        fmt_chars = 0
        for format in options:
            if format not in self.formats:
                raise NotImplementedError(f'Style "{format}" not implemented.')
            fmt_start += self.formats[format][0]
            fmt_chars += self.formats[format][1]
        fmt_chars += self.formats["END"][1]
        return fmt_start + text + self.formats["END"][0], fmt_chars

    def prompt(self, msg):
        """Prompt the user, applying consistent formatting"""
        self.center((">> " + msg + " >>"), s=["RED"], fc=-1, end="")
        return input(" ")



import time
import copy
from .formatter import space
from .board import Board


class Game:
    """Manager for a game of IQ Tester"""

    def __init__(self, formatter, size=5):
        self.f = formatter
        self.b = Board(self.f, size)
        self.moves = None
        self.prev_states = []

    @space
    def header(self):
        self.f.center(" START NEW GAME ", fill="*")
        print()
        self.f.center("Each letter in the board represents a peg in a hole.")

    @space
    def invalid(self):
        self.f.center("* Invalid selection. Try again. *")

    def validate(self, peg):
        """Check if user input is a single letter"""
        return len(peg) == 1 and (97 <= ord(peg.lower()) < 97 + self.b.holes)

    def save_state(self):
        """Append current board to the Board object's list of `prev_boards`"""
        state = tuple(map(copy.deepcopy, [self.b.board, self.moves]))
        self.prev_states.append(state)

    @space
    def remove_one_peg(self):
        """Ask the user to remove one peg from the board to start the game"""
        peg = ""
        while True:
            peg = self.f.prompt("Pick a peg to remove to start the game")
            if self.validate(peg) and self.b.remove(peg):
                break
            self.invalid()
            print()

    @space
    def pick_peg(self):
        """Ask the user to pick a peg for the next move"""
        while True:
            pick = self.f.prompt("Choose a peg to move (or '.' to go back)")
            if pick == '.':
                self.go_back()
                continue
            if self.validate(pick):
                location = self.b.locate_peg(pick)
                if self.moves and location in self.moves:
                    return location
            # invalid pick. highlight possible choices and try again
            self.b.show(self.moves.keys())
            self.invalid()
            print()
            self.f.center("* The possible pegs to move are highlighted red. *")
            print()

    @space
    def pick_jump(self, jumps):
        """Ask the user to choose which peg they want to jump over"""
        while True:
            pick = self.f.prompt("Choose the peg you want to jump")
            if self.validate(pick):
                location = self.b.locate_peg(pick)
                for jump in jumps:
                    if jump[0] == location:
                        return jump
            self.invalid()

    def make_jump(self, pick, jump):
        """Update the board to reflect the chosen jump"""
        over, to = jump
        peg = self.b.board[pick[0]][pick[1]]
        self.b.board[pick[0]][pick[1]] = None
        self.b.board[over[0]][over[1]] = None
        self.b.board[to[0]][to[1]] = peg

    @space
    def go_back(self):
        """Revert the game to its previous state"""
        # make sure there exist previous states of the board
        if not self.prev_states:
            print()
            self.f.center("* Unable to go back *")
            print()
            return
        # pop prior state and update board and moves, then show updated board
        self.b.board, self.moves = self.prev_states.pop()
        self.b.show()

    @space
    def game_over(self):
        """Handle the end of the game and display results"""
        self.f.center(" GAME OVER ", s=["BOLD"], fill="*")
        print()
        left = self.b.pegs_left()
        match left:
            case 1:
                points = 50
                result = "1 peg left. Wow! GENIUS!! 50 points!!"
            case 2:
                points = 25
                result = "2 pegs left. Above average! 25 points!"
            case 3:
                points = 10
                result = "3 pegs left. Just so-so. 10 points."
            case _:
                points = 0
                result = f"{left} pegs left. Not good. 0 points."
        self.f.center(f"{result}", s=["GREEN"])
        print()
        print(("*" * self.f.w))
        time.sleep(1)
        return points

    @space
    def play(self):
        """Drive game sequence"""
        self.header()
        self.b.show()
        self.remove_one_peg()
        while True:
            self.moves = self.b.get_moves()
            # game is over if no possible moves
            if len(self.moves) == 0:
                self.b.show()
                return self.game_over()
            self.b.show()
            pick = self.pick_peg()
            jumps = self.moves[pick]
            if len(jumps) == 1:
                jump = jumps[0]
            else:
                jumpees = set([jump[0] for jump in jumps])
                self.b.show(jumpees, color="GREEN")
                jump = self.pick_jump(jumps)
            self.save_state()
            self.make_jump(pick, jump)
